import sys
from urllib.parse import urlsplit
import boto3
import datetime
from botocore.client import Config
import logging
import botocore.vendored.requests.exceptions as botoRequestsExceptions
import botocore.exceptions as botoExceptions

_logger = logging.getLogger('storage_stats')

class S3StorageShare:
    def __init__(self, storage_share):

        self.plugin_settings = storage_share['plugin_settings']

        self.plugin_settings.update(
            {
                'url': storage_share['url']
            }
        )

        _url = urlsplit(storage_share['url'])
        self.uri = {
            'netloc':   _url.netloc,
            'path':     _url.path,
            'scheme':   _url.scheme,
        }

        self.validators = {
            'conn_timeout': {
                'default': 10,
                'required': False,
                'status_code': '005',
                'type': 'int',
            },
            'ssl_check': {
                'boolean': True,
                'default': True,
                'required': False,
                'status_code': '006',
                'valid': ['true', 'false', 'yes', 'no']
            },
            's3.priv_key': {
                'required': True,
                'status_code': '021',
            },
            's3.pub_key': {
                'required': True,
                'status_code': '022',
            },
            's3.region': {
                'default': 'us-east-1',
                'required': False,
                'status_code': '023',
            },
            's3.signature_ver': {
                'default': 's3v4',
                'required': False,
                'status_code': '024',
                'valid': ['s3', 's3v4'],
            },
        }

        self.validate_plugin_settings()

        self.uri['bucket'] = self.uri['path'].rpartition('/')[-1]


    def validate_plugin_settings(self):
        """ 
        checks for required plugin settings
        """

        _logger.info("Validating configured setitngs.")

        for _setting in self.validators:
            _logger.debug(f"Validating setting {_setting}")

            try:
                self.plugin_settings[_setting]

            except KeyError:
                if self.validators[_setting]['required']:
                    _logger.error(f"Missing required setting: {_setting}")

                else:
                    _logger.warn(f"Missing setting: {_setting}, using default value: {self.validators[_setting]['default']}")
                    self.plugin_settings.update({_setting: self.validators[_setting]['default']})

            else:
                try:
                    if self.plugin_settings[_setting].lower() not in self.validators[_setting]['valid']:
                        _logger.error(f"Invalid setting {self.plugin_settings[_setting]} for {_setting}. \
                            Valid settings: {self.validators[_setting]['valid']}")
                        exit(self.validators[_setting]['status_code'])

                    else:
                        try:
                            self.validators[_setting]['boolean']
                        except KeyError:
                            pass
                        else:
                            self.plugin_settings[_setting] = self.plugin_settings[_setting].lower() == 'true' \
                                or self.plugin_settings[_setting].lower() == 'yes'
   
                except KeyError:
                    pass


    def get_object_checksum(self, hash_type, object_url):
        """
        return the metadata of the specified hash type.
        returns None if the metadata does not exist
        """
        _metadata = self.get_object_metadata(object_url)
        
        try:
            _logger.info(f"Checking if metadata contains checksum of hash type {hash_type}")
            _logger.debug(f"Metadata being checked: {_metadata}")
            return _metadata[hash_type]
        
        except KeyError:
            _logger.warning(f"No checksum found for hash type {hash_type}")
            return None


    def get_object_metadata(self, object_url):
        """
        get and return the metadata from the file
        in s3 storage
        """
        _connection = self.get_s3_boto_client()
        _kwargs = {
            'Bucket': self.uri['bucket'],
            'Key': object_url,
        }

        _result = run_boto_client(_connection, 'head_object', _kwargs)

        try:
            _metadata = {k.lower(): v for k, v in _result['Metadata'].items()}
            return _metadata

        except KeyError:
            _logger.warning("No key Metadata in _result")
            return {}


    def put_object_checksum(self, checksum, hash_type, object_url, force):
        """
        check if there is existing metadata in the s3 file metadata
        if metadata exists, exit unless force is true
        """
        _metadata = self.get_object_metadata(object_url)

        if hash_type not in _metadata:
            # no metadata exists of this hash type, add metadata
            _metadata.setdefault(hash_type, checksum)
            _logger.info(f"New metadata detected, calling API to upload: {_metadata}")
            
            self.put_object_metadata(_metadata, object_url)

        elif force:
            # metadata already exists, replace metadata
            _metadata[hash_type] = checksum
            _logger.info(f"Force flag detected, calling API to update: {_metadata}")
            
            self.put_object_metadata(_metadata, object_url)

        else:
            # metadata already exists
            _logger.info("No new metadata detected, no need to call API. To update the existing data, use --force")
            _logger.debug(f"Metadata: {_metadata}")
            exit(0)


    def put_object_metadata(self, metadata, object_url):
        """
        put the new data in the metadata of the
        specified file in s3 storage
        """
        _connection = self.get_s3_boto_client()

        _kwargs = {
            'Bucket': self.uri['bucket'],
            'CopySource': {
                'Bucket': self.uri['bucket'],
                'Key': object_url,
            },
            'Key': object_url,
            'Metadata': metadata,
            'MetadataDirective': 'REPLACE',
        }

        try:
            # ensure there is new metadata to enter
            assert len(metadata) != 0

            _logger.info(f"Updating metdata of object '{object_url}'")
            _logger.debug(f"Metadata being uploaded: '{metadata}'")

            run_boto_client(_connection, 'copy_object', _kwargs)

        except AssertionError as INFO:
            _logger.info(f"Empty metadata. Skipping API request. {INFO}")
            sys.exit(1)


    def get_s3_boto_client(self):
        """
        create and return a connection to s3 storage
        """
        _api_url = f"{self.uri['scheme']}://{self.uri['netloc']}"
        _session = boto3.session.Session()

        _connection = _session.client(
            "s3",
            region_name=self.plugin_settings['s3.region'],
            endpoint_url=_api_url,
            aws_access_key_id=self.plugin_settings['s3.pub_key'],
            aws_secret_access_key=self.plugin_settings['s3.priv_key'],
            use_ssl=True,
            verify=self.plugin_settings['ssl_check'],
            config = Config(
                signature_version=self.plugin_settings['s3.signature_ver'],
                connect_timeout=int(self.plugin_settings['conn_timeout']),
                retries=dict(max_attempts=0)
            )
        )

        return _connection


    def list_objects(self, prefix):
        """
        get the files in s3 storage, calculate, and 
        return the number of files and bytes used
        """
        _connection = self.get_s3_boto_client()

        _total_bytes = 0
        _total_files = 0

        _kwargs = {
            'Bucket': self.uri['bucket'],
            'Prefix': prefix,
        }

        while True:
            _response = run_boto_client(_connection, 'list_objects', _kwargs)

            # check for files in the response
            try:
                _response['Contents']
            except KeyError:
                break
            else:
                for _file in _response['Contents']:
                    _total_bytes += int(_file['Size'])
                    _total_files += 1

            # marker indicating the location of the last file read        
            try:
                _kwargs['Marker'] = _response['NextMarker']
            except KeyError:
                break

        return int(_total_bytes), _total_files


def run_boto_client(_connection, method, _kwargs):
    """
    runs the specified function with the 
    given arguments and returns the result
    """
    _function = getattr(_connection, method)
    _result = {}
    try:
        _result = _function(**_kwargs)

    except botoExceptions.ClientError as ERR:
        _logger.error(f"[{ERR.__class__.__name__}][{ERR.response['ResponseMetadata']['HTTPStatusCode']}] Failed to establish a connection")
        _logger.debug(f"{str(ERR)}")
        print("An error occurred (404). File not found")
        sys.exit(1)

    except botoRequestsExceptions.SSLError as ERR:
        _logger.error(f"[{ERR.__class__.__name__}][092] Failed to establish a connection")
        _logger.debug(f"{str(ERR)}")

    except botoRequestsExceptions.RequestException as ERR:
        _logger.error(f"[{ERR.__class__.__name__}][400] Failed to establish a connection")
        _logger.debug(f"{str(ERR)}")

    except botoExceptions.ParamValidationError as ERR:
        _logger.error(f"[{ERR.__class__.__name__}][095] Failed to establish a connection")
        _logger.debug(f"{str(ERR)}")

    except botoExceptions.BotoCoreError as ERR:
        _logger.error(f"[{ERR.__class__.__name__}][400] Failed to establish a connection")
        _logger.debug(f"{str(ERR)}")

    return _result



import argparse
import sys

def parse_args(args):
    """
    create the argument parser
    """
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()

    add_checksums_subparser(subparser)
    add_reports_subparser(subparser)

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args(args)


def add_checksums_subparser(subparser):
    """
    create the checksums subparser
    """
    parser = subparser.add_parser(
        'checksums',
        help="obtain and output object/file checksums"
    )
    subparser = parser.add_subparsers()
    parser.set_defaults(cmd='checksums')

    add_checksums_get_subparser(subparser)
    add_checksums_put_subparser(subparser)


def add_checksums_get_subparser(subparser):
    """
    create the checksums get subparser
    """
    parser = subparser.add_parser(
        'get', 
        help="Get object/file checksums."
    )
    parser.set_defaults(sub_cmd='get')

    add_checksum_options(parser)


def add_checksums_put_subparser(subparser):
    """
    create the checksums put subparser
    """
    parser = subparser.add_parser(
        'put',
        help="Set object/file checksums."
    )

    parser.set_defaults(sub_cmd='put')

    group_checksum = add_checksum_options(parser)

    group_checksum.add_argument(
        '--checksum',
        action='store',
        default=False,
        required=True,
        dest='checksum',
        type=str.lower,
        help="String with checksum to set. "
            "Required."
    )


def add_reports_subparser(subparser):
    """
    create the reports subparser
    """
    parser = subparser.add_parser(
        'reports',
        help="obtain and output storage reports"
    )
    subparser = parser.add_subparsers()
    parser.set_defaults(cmd='reports')

    add_general_options(parser)
    add_logging_options(parser)

    report_options = parser.add_argument_group("Report options")

    report_options.add_argument(
        '-p', '--prefix',
        action='store',
        default='',
        dest='prefix',
        type=str,
        help="Prefix for directory path. "
                "Default: ''"
    )


def add_checksum_options(parser):
    """
    create the common checksum options
    """
    add_general_options(parser)
    add_logging_options(parser)

    group_checksum = parser.add_argument_group("Checksum required options")

    group_checksum.add_argument(
        '-t', '--hash_type',
        action='store',
        default=False,
        required=True,
        dest='hash_type',
        type=str.lower,
        help="Type of checksum hash. ['adler32', md5] "
                "Required."
    )

    group_checksum.add_argument(
        '-f', '--file',
        action='store',
        default=False,
        required=True,
        dest='file',
        help="URL of object/file to request checksum of. "
                "Required."
    )

    return group_checksum


def add_general_options(parser):
    """
    create the common options for both
    checksums and reports
    """
    parser.add_argument(
        '-c', '--config',
        action='store',
        default='/etc/xrootd/s3cfg',
        dest='config_path',
        type=str.lower,
        help="Path to s3 endpoint .conf file or directory. "
            "Accepts one argument. "
            "Default: '/etc/xrootd/s3cfg'."
    )

    parser.add_argument(
        '--force',
        action='store_true',
        default=False,
        dest='force',
        help="Force command execution."
    )

    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        default=False,
        dest='verbose',
        help="Show on stderr events according to loglevel."
    )
    parser.add_argument(
        '-n', '--name',
        action='store',
        default='',
        dest='s3_name',
        help="Set the name of the s3 service. "
             "Default: ''"
    )



def add_logging_options(parser):
    """Add logging optional arguments.

    Arguments:
    parser -- Object form argparse.ArgumentParser()

    """

    group_logging = parser.add_argument_group("Logging options")

    group_logging.add_argument(
        '--logfile',
        action='store',
        default=False,
        dest='logfile',
        help="Set logfiles path. "
             "Default: /var/log/xrootd/[S3_NAME]/s3_storage_stats.log"
    )

    group_logging.add_argument(
        '--loglevel',
        action='store',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
        default='WARNING',
        dest='loglevel',
        help="Set log output level. "
        "Default: WARNING."
    )



from os import path
from setuptools import setup, find_packages

PWD = path.abspath(path.dirname(__file__))

with open(path.join(PWD, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
	name='s3-storage-stats',
	description='S3 Storage Stats',
	long_description=long_description,
	long_description_content_type='text/markdown',
	version='1.0.6',
	author='Alyssa Blair',
	author_email='alyssablair@uvic.ca',
	packages=['s3_storage_stats'],
	install_requires=['boto3>=1.2.8'],
	scripts=['s3-storage-stats'],
)




import logging
import logging.handlers
import os

def setup_logger(logfile, loglevel, s3_name, verbose):
    """Setup the logger format to be used throughout the script.

    Arguments:
    logfile -- string defining path to write logs to.
    loglevel -- string defining level to log: "DEBUG, INFO, WARNING, ERROR"
    verbose -- boolean. 'True' prints log messages to stderr.

    """
    # To capture warnings emitted by modules.
    logging.captureWarnings(True)

    # Create file logger.
    _logger = logging.getLogger("storage_stats")

    # Set log level to use.
    _num_loglevel = getattr(logging, loglevel.upper())
    _logger.setLevel(_num_loglevel)

    # create default value for logfile if not specified
    logfile = logfile or os.path.join("/var/log/xrootd/", s3_name, "s3_storage_stats.log")        

    # Set file where to log and the mode to use and set the format to use.
    _log_handler_file = logging.handlers.TimedRotatingFileHandler(
        logfile,
        when="midnight",
        backupCount=15,
    )

    # Set logger format
    _log_format_file = logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s')

    # Set the format to the file handler.
    _log_handler_file.setFormatter(_log_format_file)

    # Add the file handler.
    _logger.addHandler(_log_handler_file)

    # Create STDERR handler if verbose is requested and add it to logger.
    if verbose:
        log_handler_stderr = logging.StreamHandler()
        log_format_stderr = logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s')

        log_handler_stderr.setLevel(_num_loglevel)
        log_handler_stderr.setFormatter(log_format_stderr)
        # Add handler
        _logger.addHandler(log_handler_stderr)
